// Code generated by go tool dist; DO NOT EDIT.
// This is a bootstrap copy of D:\bld\go-nocgo_1587274346108\work\go\src\cmd\compile\internal\gc\norace.go

//line D:\bld\go-nocgo_1587274346108\work\go\src\cmd\compile\internal\gc\norace.go:1
// Copyright 2017 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

// +build !race

package gc

const raceEnabled = false
